<?php
namespace Ep\TecDoc;

use Ep\TecDoc\Exception\HttpException;
use Exception;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\GuzzleException;
use Psr\Cache\CacheItemPoolInterface;
use Laminas\Db\TableGateway\TableGateway;

/**
 * TecDoc Servis
 *
 * Kroz ovu klasu se dohvaćaju svi podaci iz TecDoc-a.
 *
 * @package Ep\TecDoc
 */
class TecDocService
{
    const SERVICE_URI = 'https://webservice.tecalliance.services';
    const SERVICE_ENDPOINT = '/pegasus-3-0/services/TecdocToCatDLB.jsonEndpoint';

	/** @var array */
	protected $config;
	/** @var ClientInterface */
	protected $client;
	/** @var CacheItemPoolInterface */
	protected $cache;
	/** @var array */
	protected $selectedVehicle;
	/** @var TableGateway */
	private $attributesTable;

    /**
     * Konstruktor.
     *
     * @param array $config
     * @param ClientInterface $client
     * @param CacheItemPoolInterface $cache
     *
     * @throws Exception
     */
    public function __construct(
        array $config = [],
        ClientInterface $client,
        CacheItemPoolInterface $cache,
        TableGateway $attributesTable = null
    ) {
	    $this->config = $config;
	    $this->client = $client;
	    $this->cache = $cache;
	    $this->attributesTable = $attributesTable;

	    $this->allowClientAccess();
    }

    /**
     * Daje pristup klijentovom IP-u TecDocu-.
     *
     * @throws Exception
     */
    protected function allowClientAccess()
    {
        if (!isset($_SERVER['REMOTE_ADDR'])) {
            return;
        }

    	if (!isset($_SESSION['tecdoc_allow']) || (time() - $_SESSION['tecdoc_allow']) > 23 * 3600 || $_SESSION['tecdoc_allow_ip'] != $_SERVER['REMOTE_ADDR']) {
    		if (!in_array($_SERVER['REMOTE_ADDR'], ['127.0.0.1', '::1'])) {
    			if (!isset($_SESSION['tecdoc_allow_ip']) || $_SESSION['tecdoc_allow_ip'] != $_SERVER['REMOTE_ADDR']) {
				    $this->request('addDynamicAddress', [
					    'address'       => $_SERVER['REMOTE_ADDR'],
					    'validityHours' => 24,
				    ]);

				    $_SESSION['tecdoc_allow'] = time();
				    $_SESSION['tecdoc_allow_ip'] = $_SERVER['REMOTE_ADDR'];
			    }
		    }
	    }
    }

	/**
     * Upućuje request prema TecDoc servisu.
     *
	 * @param string $method TecDoc metoda.
	 * @param array $params Parametri requesta.
	 *
	 * @return array|mixed
	 * @throws Exception
	 */
    public function request($method, array $params = [])
    {
	    $params = [$method => $this->checkParams($params)];

        try {
            $response = $this->client->request('POST', self::SERVICE_ENDPOINT, [
                'json' => $params,
		    ]);
	    } catch (GuzzleException $e) {
		    throw new HttpException(sprintf('TecDoc HTTP Greška: %s', $e->getMessage()), 0, $e);
	    }

	    if ($response->getStatusCode() > 200) {
		    throw new HttpException(sprintf(
			    'TecDoc HTTP Greška: %d, %s',
			    $response->getStatusCode(),
			    $response->getReasonPhrase()
		    ));
	    }

	    $jsonContent = (string) $response->getBody();
	    $responseData = json_decode($jsonContent, true);

        if (null === $responseData) {
        	throw new \RuntimeException(sprintf(
        		"Došlo je do pogreške prilikom dekodiranja JSON-a: %s.\nJSON: %s",
		        json_last_error_msg(),
		        $jsonContent
	        ));
        }

	    if ($responseData['status'] > 200) {
		    throw new HttpException(sprintf(
			    'TecDoc HTTP Greška: %d, %s',
			    $responseData['status'],
			    $responseData['statusText']
		    ));
	    }

	    return isset($responseData['data']['array']) && is_array($responseData['data']['array'])
		    ? $responseData['data']['array']
		    : [];
    }

    /**
     * Dodaje zajedničke parametre iz konfiguracije u request.
     *
     * @param array $params Parametri requesta.
     *
     * @return array
     */
    protected function checkParams(array $params)
    {
    	if (!isset($params['linkingTargetType'])) {
    		$params['linkingTargetType'] = $this->config['linkingTargetType'];
	    }

	    if (!isset($params['carType'])) {
    		$params['carType'] = $this->config['linkingTargetType'];
	    }

	    if (!isset($params['lang'])) {
    		$params['lang'] = $this->config['language'];
	    }

	    if (!isset($params['country'])) {
    		$params['country'] = $this->config['country'];
	    }

	    if (!isset($params['countriesCarSelection'])) {
    		$params['countriesCarSelection'] = $this->config['country'];
	    }

	    if (!isset($params['articleCountry'])) {
    		$params['articleCountry'] = $this->config['country'];
	    }

	    if (!isset($params['provider'])) {
    		$params['provider'] = $this->config['provider'];
	    }

	    return $params;
    }

    /**
     * Vraća trenutno odabrano vozilo.
     *
     * @return mixed[]|null
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function getSelectedVehicle()
    {
    	if (null === $this->selectedVehicle) {
            if (isset($_COOKIE['_tv'])) {
                $this->selectedVehicle = $this->getVehicle($_COOKIE['_tv']);
            } else if (isset($_COOKIE['_tecdoc2'])) {
                $this->selectedVehicle = $this->getVehicle($_COOKIE['_tecdoc2']);
            }
        }

	    return $this->selectedVehicle;
    }

    /**
     * Postavlja trenutno odabrano vozilo.
     *
     * @param string|array $vehicle
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function setSelectedVehicle($vehicle)
    {
    	if (!is_array($vehicle)) {
    		$vehicle = $this->getVehicle($vehicle);
	    }

	    if (!$vehicle) {
    	    return;
        }

	    $current = $this->getSelectedVehicle();

    	if (!$current || $current['id'] != $vehicle['id']) {
    		$this->selectedVehicle = $vehicle;
	    }
    }

    //<editor-fold desc="Proizvođači vozila">

    /**
     * Vraća proizvođače vozila.
     *
     * @param string|null $country
     * @param bool $countryGroupFlag
     * @param int|null $favouredList
     * @param string|null $lang
     * @param bool $returnRaw
     *
     * @return array[]
     *
     * @throws \Psr\Cache\InvalidArgumentException
     * @throws Exception
     */
    public function getManufacturers(
        string $country = null,
        bool $countryGroupFlag = false,
        int $favouredList = null,
        string $lang = null,
        bool $returnRaw = false
    ) {
    	$item = $this->cache->getItem('manufacturers.old.all');

    	if ($returnRaw || !$item->isHit()) {
    	    $params = [
    	        'country' => $country,
                'countryGroupFlag' => $countryGroupFlag,
                'lang' => $lang,
            ];

    	    if (null !== $favouredList) {
    	        $params['favouredList'] = $favouredList;
            }

		    $manufacturers = $this->request('getManufacturers', $params);
		    $config = $this->config;

		    $manufacturers = array_filter($manufacturers, function($manufacturer) use ($config) {
			    return !in_array($manufacturer['manuId'], $config['excludedManufacturerIds']);
		    });

		    if ($returnRaw) {
		        return $manufacturers;
            }

		    $manufacturers = array_map(function($manufacturer) {
		    	if ('SKODA' === $manufacturer['manuName']) {
		    		$manufacturer['manuName'] = 'ŠKODA';
			    }

		    	return [
				    'id' => $manufacturer['manuId'],
				    'name' => $manufacturer['manuName'],
				    'slug' => sanitize_title_with_dashes(remove_accents($manufacturer['manuName'])),
			    ];
		    }, $manufacturers);

		    $item->set($manufacturers);
		    $this->cache->save($item);
	    }

	    return $item->get();
    }

    /**
     * Vraća izdvojene proizvođače vozila.
     *
     * @return array[]
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function getFeaturedManufacturers()
    {
	    $item = $this->cache->getItem('manufacturers.featured');

	    if (!$item->isHit()) {
	    	$config = $this->config;
	    	$featured = [];

	    	foreach ($config['featuredManufacturerIds'] as $id) {
	    	    $featured[] = $this->getManufacturer($id);
            }

	    	$item->set($featured);
	    	$this->cache->save($item);
	    }

	    return $item->get();
    }

    /**
     * Vraća proizvođača po id-u ili slug-u.
     *
     * @param int|string $idOrSlug
     *
     * @return mixed
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
	public function getManufacturer($idOrSlug)
	{
		$field = is_numeric($idOrSlug) ? 'id' : 'slug';
		$cacheKey = 'id' === $field
			? sprintf('manufacturers.%s.%d', $field, intval($idOrSlug))
			: sprintf('manufacturers.%s.%s', $field, md5($idOrSlug));
		$item = $this->cache->getItem($cacheKey);

		if (!$item->isHit()) {
			foreach ($this->getManufacturers() as $manufacturer) {
				if ($manufacturer[$field] == $idOrSlug) {
					$item->set($manufacturer);
					$this->cache->save($item);
					break;
				}
			}
		}

		return $item->get();
	}

	//</editor-fold>
	//<editor-fold desc="Modeli vozila">

    /**
     * Vraća modele vozila za datog proizvođača.
     *
     * @param int|array $manufacturer
     * @param string|null $country
     * @param string|null $lang
     * @param int|null $favouredList
     * @param bool $returnRaw
     *
     * @return array[]
     *
     * @throws \Psr\Cache\InvalidArgumentException
     * @throws Exception
     */
    public function getModels(
        $manufacturer,
        string $country = null,
        string $lang = null,
        int $favouredList = null,
        bool $returnRaw = false
    ) {
    	if (!is_array($manufacturer)) {
		    $manufacturer = $this->getManufacturer($manufacturer);
	    }

    	$cacheKey = sprintf('models.%d', $manufacturer['id']);
    	$item = $this->cache->getItem($cacheKey);

    	if ($returnRaw || !$item->isHit()) {
            $params = [
                'manuId' => $manufacturer['id'],
                'country' => $country,
                'lang' => $lang,
            ];

            if (null !== $favouredList) {
                $params['favouredList'] = $favouredList;
            }

		    $raw = $this->request('getModelSeries', $params);

            if ($returnRaw) {
                return $raw;
            }

		    $models = [];

		    foreach ($raw as $model) {
		    	$yearFrom = isset($model['yearOfConstrFrom'])
			        ? $this->formatConstructionYear($model['yearOfConstrFrom'])
			        : '...';

		    	$yearTo = isset($model['yearOfConstrTo'])
			        ? $this->formatConstructionYear($model['yearOfConstrTo'])
			        : '...';

		    	$year = sprintf('%s - %s', $yearFrom, $yearTo);
		    	$modelSlug = sprintf('%s-%d', sanitize_title_with_dashes(remove_accents($model['modelname'])), $model['modelId']);
		    	$slug = sprintf('%s--%s', $manufacturer['slug'], $modelSlug);

		    	$models[] = [
				    'id' => $model['modelId'],
				    'manufacturerId' => $manufacturer,
				    'name' => $model['modelname'],
				    'slug' => $slug,
				    'year_from' => $yearFrom,
				    'year_to' => $yearTo,
				    'year' => $year,
			    ];
		    }

		    $item->set($models);
		    $this->cache->save($item);
	    }

	    return $item->get();
    }

    /**
     * Vraća model po slug-u.
     *
     * @param string $slug
     *
     * @return string[]
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
	public function getModel($slug)
	{
		$cacheKey = sprintf('models.slug.%s', md5($slug));
		$item = $this->cache->getItem($cacheKey);

		if (!$item->isHit()) {
			$slugParts = explode('--', $slug);

			if (count($slugParts) !== 2) {
				throw new \RuntimeException(sprintf(
					"Neispravan slug modela '%s'.", $slug
				));
			}

			$manufacturer = $this->getManufacturer($slugParts[0]);

			foreach ($this->getModels($manufacturer) as $model) {
				if ($slug === $model['slug']) {
					$item->set($model);
					$this->cache->save($item);
					break;
				}
			}
		}

		return $item->get();
	}

    /**
     * Vraća model po id-u.
     *
     * @param $manufacturerId
     * @param $modelId
     *
     * @return string[]
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
    public function getModelById($manufacturerId, $modelId)
    {
    	$cacheKey = sprintf('models.id.%d', $modelId);
    	$item = $this->cache->getItem($cacheKey);

    	if (!$item->isHit()) {
		    foreach ($this->getModels($manufacturerId) as $model) {
		    	if ($modelId == $model['id']) {
		    		$item->set($model);
		    		$this->cache->save($item);
		    		break;
			    }
		    }
	    }

	    return $item->get();
    }

    //</editor-fold>
	//<editor-fold desc="Vozila">

    /**
     * Vraća vozila po proizvođaču i modelu.
     *
     * @param int $manufacturerId Id proizvođača.
     * @param int $modelId Id modela.
     *
     * @return array[]
     *
     * @throws \Psr\Cache\InvalidArgumentException
     * @throws Exception
     */
    public function getVehicles(
        $manufacturerId,
        $modelId
    ) {
    	$cacheKey = sprintf('vehicles.%d.%d', $manufacturerId, $modelId);
    	$item = $this->cache->getItem($cacheKey);

    	if (!$item->isHit()) {
		    $vehicleIds = $this->request('getVehicleIdsByCriteria', [
			    'manuId' => intval($manufacturerId),
			    'modId' => intval($modelId),
		    ]);

		    $vehicleIds = array_map(function($vehicle) {
			    return $vehicle['carId'];
		    }, $vehicleIds);

		    $vehicles = $this->getVehiclesByIds($vehicleIds);

		    $item->set($vehicles);
	    }

	    return $item->get();
    }

    /**
     * @param int $manuId
     * @param int $modId
     * @param string $countriesCarSelection
     * @param string $lang
     * @param bool $countryGroupFlag
     *
     * @return array|mixed
     * @throws Exception
     */
    public function getVehiclesByCriteria(
        int $manuId,
        int $modId,
        string $countriesCarSelection,
        string $lang,
        bool $countryGroupFlag = false
    ) {
        $carIds = $this->request('getVehicleIdsByCriteria', [
            'manuId' => $manuId,
            'modId' => $modId,
            'countriesCarSelection' => $countriesCarSelection,
            'countryGroupFlag' => $countryGroupFlag,
        ]);

        $carIds = array_map(function($car) {
            return $car['carId'];
        }, $carIds);

        if (count($carIds) === 0) {
            return [];
        }

        $vehicles = $this->request('getVehicleByIds4', [
            'articleCountry' => $countriesCarSelection,
            'carIds' => ['array' => $carIds],
            'countriesCarSelection' => $countriesCarSelection,
            'country' => $countriesCarSelection,
            'countryGroupFlag' => $countryGroupFlag,
            'lang' => $lang,
            'motorCodes' => true,
        ]);

        foreach ($vehicles as &$vehicle) {
            $vehicle['vehicleDetails']['carName'] = $vehicle['vehicleDetails']['typeName'];
            $vehicle['vehicleDetails']['cylinderCapacity'] = $vehicle['vehicleDetails']['cylinderCapacityCcm'];
        }

        return $vehicles;
    }

    /**
     * Vraća detaljna vozila po id-evima.
     *
     * @param array $vehicleIds
     * @param $manufacturer
     * @param $model
     *
     * @return array[]
     *
     * @throws \Psr\Cache\InvalidArgumentException
     * @throws Exception
     */
	protected function getVehiclesByIds(array $vehicleIds)
	{
	    $vehicleIds = array_filter($vehicleIds);
	    if (empty($vehicleIds)) return [];

		$vehicles = [];
		$chunks = array_chunk($vehicleIds, 25);

		foreach ($chunks as $chunk) {
			$chunk = $this->request('getVehicleByIds4', [
				'axles' => true,
				'cabs' => true,
				'carIds' => ['array' => $chunk],
				'motorCodes' => true,
				'secondaryTypes' => true,
				'wheelbases' => true,
			]);

			$chunk = array_map(function($vehicle) {
				$slug = implode('--', [
				    sanitize_title_with_dashes(remove_accents($vehicle['vehicleDetails']['manuName'])),
                    sprintf('%s-%d', sanitize_title_with_dashes(remove_accents($vehicle['vehicleDetails']['modelName'])), $vehicle['vehicleDetails']['modId']),
                    sanitize_title_with_dashes(remove_accents(implode('-', [
                        $vehicle['vehicleDetails']['typeName'],
                        $vehicle['vehicleDetails']['powerKwFrom'] . 'kw',
                        $vehicle['carId'],
                    ])))
                ]);

				$image = isset($vehicle['vehicleDocId']) ? $this->getImageUrl($vehicle['vehicleDocId']) : null;

				return [
					'id' => $vehicle['carId'],
                    'group' => 'default',
					'manufacturerId' => $vehicle['vehicleDetails']['manuId'],
                    'manufacturerName' => $vehicle['vehicleDetails']['manuName'],
					'modelId' => $vehicle['vehicleDetails']['modId'],
                    'modelName' => $vehicle['vehicleDetails']['modelName'],
					'name' => $vehicle['vehicleDetails']['typeName'],
					'slug' => $slug,
                    'ccm' => $vehicle['vehicleDetails']['cylinderCapacityCcm'],
					'constructionType' => $vehicle['vehicleDetails']['constructionType'],
					'engine' => $vehicle['vehicleDetails']['motorType'],
					'power' => $vehicle['vehicleDetails']['powerKwFrom'] . ' kw',
					'powerHp' => $vehicle['vehicleDetails']['powerHpFrom'] . ' ks',
					'image' => $image,
				];
			}, $chunk);

			$vehicles = array_merge($vehicles, $chunk);
		}

		return $vehicles;
	}

    /**
     * Vraća vozilo po slug-u.
     *
     * @param string $slug
     *
     * @return mixed[]
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
	public function getVehicle($slug)
	{
		$cacheKey = sprintf('vehicles.slug.%s', md5($slug));
		$item = $this->cache->getItem($cacheKey);

		if (!$item->isHit()) {
		    if (is_numeric($slug)) {
		        $vehicleId = $slug;
            } else {
                $slugParts = explode('--', $slug);

                if (count($slugParts) !== 3) {
                    return null;
                }

                preg_match('/-(\d+)$/', $slug, $matches);
                $vehicleId = $matches[1];
            }

            if ($vehicle = $this->getVehicleById($vehicleId)) {
		        $item->set($vehicle);
		        $this->cache->save($item);
            }
		}

		return $item->get();
	}

    /**
     * @param $manufacturerId
     * @param $modelId
     * @param $vehicleId
     *
     * @return mixed
     *
     * @throws \Psr\Cache\InvalidArgumentException
     */
	public function getVehicleById($vehicleId)
	{
		$cacheKey = sprintf('vehicles.id.old.%d', $vehicleId);
		$item = $this->cache->getItem($cacheKey);

		if (!$item->isHit()) {
		    $vehicles = $this->getVehiclesByIds([$vehicleId]);

		    if (count($vehicles) > 0) {
		        $item->set($vehicles[0]);
		        $this->cache->save($item);
            }
		}

		return $item->get();
	}

    /**
     * @param int $articleId
     * @param string|null $linkingTargetType
     * @param string|null $articleCountry
     * @param string|null $lang
     *
     * @return array|mixed
     * @throws Exception
     */
	public function getArticleLinkedAllLinkingTarget(
	    int $articleId,
        string $linkingTargetType = null,
        string $articleCountry = null,
        string $lang = null
    ) {
        return $this->request('getArticleLinkedAllLinkingTarget4', [
            'articleId' => $articleId,
            'articleCountry' => $articleCountry,
            'lang' => $lang,
            'linkingTargetType' => $linkingTargetType,
        ]);
    }

    /**
     * @param int $articleId
     * @param array $articleLinks
     * @param string|null $linkingTargetType
     * @param string|null $articleCountry
     * @param string|null $lang
     * @param bool $immediateAttributs
     *
     * @return array|mixed
     * @throws Exception
     */
    public function getArticleLinkedAllLinkingTargetsByIds(
        int $articleId,
        array $articleLinks,
        string $linkingTargetType = null,
        string $articleCountry = null,
        string $lang = null,
        bool $immediateAttributs = false
    ) {
        return $this->request('getArticleLinkedAllLinkingTargetsByIds3', [
            'articleCountry' => $articleCountry,
            'articleId' => $articleId,
            'lang' => $lang,
            'linkedArticlePairs' => [
                'array' => $articleLinks,
            ],
            'linkingTargetType' => $linkingTargetType,
            'immediateAttributs' => $immediateAttributs,
        ]);
    }

    /**
     * @param $articleId
     *
     * @return mixed
     *
     * @throws \Psr\Cache\InvalidArgumentException
     * @throws Exception
     */
	public function getVehicleIdsLinkedToArticle($articleId) {
		$cacheKey = sprintf('vehicles.article_id.vehicles.%d', $articleId);
		$item = $this->cache->getItem($cacheKey);

		if (!$item->isHit()) {
			$vehicleIds = [];

			$articleLinks = $this->getArticleLinkedAllLinkingTarget($articleId);

			if (isset($articleLinks[0]['articleLinkages']['array'])) {
				$linkedArticlePairs = array_map(function($link) {
					return [
						'articleLinkId' => $link['articleLinkId'],
						'linkingTargetId' => $link['linkingTargetId'],
					];
				}, $articleLinks[0]['articleLinkages']['array']);

				$chunks = array_chunk($linkedArticlePairs, 25);

				foreach ($chunks as $chunk) {
					$vehicles = $this->request('getArticleLinkedAllLinkingTargetsByIds3', [
						'articleId' => $articleId,
						'immediateAttributs' => true,
						'linkedArticlePairs' => ['array' => $chunk],
					]);

					$vehiclesArray = array_map(function($vehicle) {
						return $vehicle['linkedVehicles']['array'];
					}, $vehicles);

					foreach ($vehiclesArray as $_vehicles) {
						foreach ($_vehicles as $vehicle) {
							$vehicleIds[] = $vehicle['carId'];
						}
					}
				}
			}

			$item->set($vehicleIds);
			$this->cache->save($item);
		}

		return $item->get();
	}

    /**
     * @param $articleId
     *
     * @return mixed
     *
     * @throws \Psr\Cache\InvalidArgumentException
     * @throws Exception
     */
	public function getVehicleStringsLinkedToArticle($articleId)
	{
		$cacheKey = sprintf('vehicles.article_id.%d', $articleId);
		$item = $this->cache->getItem($cacheKey);

		if (!$item->isHit()) {
			$allVehicles = [];
			$articleLinks = $this->request('getArticleLinkedAllLinkingTarget4', [
				'articleId' => $articleId,
			]);

			if (isset($articleLinks[0]['articleLinkages']['array'])) {
                if (count($articleLinks[0]['articleLinkages']['array']) > 5000) {
                    $articleLinks[0]['articleLinkages']['array'] = array_slice($articleLinks[0]['articleLinkages']['array'], 0, 5000);
                }

				$linkedArticlePairs = array_map(function($link) {
					return [
						'articleLinkId' => $link['articleLinkId'],
						'linkingTargetId' => $link['linkingTargetId'],
					];
				}, $articleLinks[0]['articleLinkages']['array']);

				$chunks = array_chunk($linkedArticlePairs, 25);

				foreach ($chunks as $chunk) {
					$vehicles = $this->request('getArticleLinkedAllLinkingTargetsByIds3', [
						'articleId' => $articleId,
						'immediateAttributs' => true,
						'linkedArticlePairs' => ['array' => $chunk],
					]);

					$vehiclesArray = array_map(function($vehicle) {
						return $vehicle['linkedVehicles']['array'];
					}, $vehicles);

					foreach ($vehiclesArray as $_vehicles) {
						foreach ($_vehicles as $vehicle) {
							if (!isset($allVehicles[$vehicle['manuDesc']])) {
								$allVehicles[$vehicle['manuDesc']] = [];
							}

                            $slug = implode('--', [
                                sanitize_title_with_dashes(remove_accents($vehicle['manuDesc'])),
                                sprintf('%s-%d', sanitize_title_with_dashes(remove_accents($vehicle['modelDesc'])), $vehicle['modelId']),
                                sanitize_title_with_dashes(remove_accents(implode('-', [
                                    $vehicle['carDesc'],
                                    $vehicle['powerKwFrom'] . 'kw',
                                    $vehicle['carId'],
                                ])))
                            ]);



							$allVehicles[$vehicle['manuDesc']][] = [
							    'label' => sprintf(
								'%s %s %s %s %d kw (%s - %s)',
                                    $vehicle['manuDesc'],
                                    $vehicle['modelDesc'],
                                    $vehicle['constructionType'],
                                    $vehicle['carDesc'],
                                    $vehicle['powerKwFrom'],
                                    isset($vehicle['yearOfConstructionFrom'])
                                        ? $this->formatConstructionYear($vehicle['yearOfConstructionFrom'])
                                        : '...',
                                    isset($vehicle['yearOfConstructionTo'])
                                        ? $this->formatConstructionYear($vehicle['yearOfConstructionTo'])
                                        : '...'
                                ),
                                'slug' => $slug,
                            ];
						}
					}
				}
			}

			$item->set($allVehicles);
			$this->cache->save($item);
		}

		return $item->get();
	}

	//</editor-fold>
	//<editor-fold desc="Artikli">

    /**
     * @param array $articleLinks
     *
     * @return array|mixed
     *
     * @throws \Psr\Cache\InvalidArgumentException
     * @throws Exception
     */
	public function getArticlesByLinks(array $articleLinks)
	{
		if (!($vehicle = $this->getSelectedVehicle())) {
			return [];
		}

		$cacheKey = sprintf('articles.links.%s', md5(serialize($articleLinks)));
		$item = $this->cache->getItem($cacheKey);

		if (!$item->isHit()) {
			$articles = [];
			$chunks = array_chunk($articleLinks, 25);

			foreach ($chunks as $chunk) {
				$chunk = $this->request('getAssignedArticlesByIds7', [
					'articleIdPairs' => ['array' => $chunk],
					'linkingTargetId' => $vehicle['id'],
					'manuId' => $vehicle['manufacturerId'],
					'modId' => $vehicle['modelId'],
					'attributs' => true,
					'basicData' => true,
					'documents' => true,
					'eanNumbers' => true,
					'immediateAttributs' => true,
					'immediateInfo' => true,
					'info' => true,
					'mainArticles' => true,
					'oeNumbers' => true,
					'replacedByNumbers' => true,
					'replacedNumbers' => true,
					'thumbnails' => true,
					'usageNumbers' => true,
				]);

				$articles = array_merge($articles, $chunk);
			}

			$articles = array_map([$this, 'mapArticle'], $articles);

			$item->set($articles);
			$this->cache->save($item);
		}

		return $item->get();
	}

    /**
     * @param string|null $lang
     *
     * @return array|mixed
     * @throws Exception Kada ne možemo uredno odraditi request.
     */
	public function getLanguages(string $lang = null)
    {
        return $this->request('getLanguages', [
            'lang' => $lang,
        ]);
    }

    /**
     * @param array $articleIds
     *
     * @param bool $oeNumbers
     * @param string|null $country
     * @param string|null $lang
     * @param bool $returnArticles
     *
     * @return mixed
     *
     * @throws \Psr\Cache\InvalidArgumentException
     * @throws Exception
     */
	public function getArticlesByIds(
	    array $articleIds,
        bool $oeNumbers = true,
        string $country = null,
        string $lang = null,
        bool $returnArticles = false
    ) {
		sort($articleIds);
		$cacheKey = sprintf('articles.ids.%s', md5(implode(',', $articleIds)));
		$item = $this->cache->getItem($cacheKey);

		if ($returnArticles || !$item->isHit()) {
			$articles = [];
			$chunks = array_chunk($articleIds, 25);

			foreach ($chunks as $chunk) {
				$chunk = $this->request('getDirectArticlesByIds4', [
					'articleId' => ['array' => $chunk],
					'attributs' => true,
					'basicData' => true,
					'documents' => true,
					'eanNumbers' => true,
					'immediateAttributs' => true,
					'immediateInfo' => true,
					'info' => true,
					'mainArticles' => true,
					'normalAustauschPrice' => false,
					'oeNumbers' => $oeNumbers,
					'prices' => false,
					'replacedByNumbers' => true,
					'replacedNumbers' => true,
					'thumbnails' => true,
					'usageNumbers' => true,

				]);

				$articles = array_merge($articles, $chunk);
			}

			if ($returnArticles) {
			    return $articles;
            }

			$articles = array_map([$this, 'mapArticle'], $articles);

			$item->set($articles);
			$this->cache->save($item);
		}

		return $item->get();
	}

	public function mapArticle($article)
	{
		if (!isset($article['directArticle']) && isset($article['assignedArticle'])) {
			$article['directArticle'] = $article['assignedArticle'];
		}

		$images = isset($article['articleDocuments']['array'])
			? array_filter(array_map(function($document) {
			    if (!in_array(intval($document['docTypeId']), [1, 3, 5])) {
			        return false;
                }

				return preg_match('/\.pdf$/i', $document['docFileName'])
					? false
                    : $this->getImageUrl($document['docId']);
			}, $article['articleDocuments']['array']))
			: [];
		$images = array_reverse(array_reverse($images));

		$fittingPositions = [];

		$attributes = isset($article['articleAttributes']['array'])
			? array_map(function($attribute) use ($article, &$fittingPositions) {
                if ($attribute['attrId'] === 1012) {
                    return null;
                }

				$value = isset($attribute['attrValue']) ? $attribute['attrValue'] : '';
				$unit = isset($attribute['attrUnit']) ? $attribute['attrUnit'] : '';

				$mapped = [
					'id' => $attribute['attrId'],
					'name' => $attribute['attrName'],
					'value' => trim(sprintf('%s %s', $value, $unit)),
				];

				if (100 === $attribute['attrId']) {
					$fittingPositions[] = $mapped;
				}

				return $mapped;
			}, $article['articleAttributes']['array'])
			: [];

		$attributes = array_merge($fittingPositions, $attributes);
        $attributes = array_filter($attributes);

		$infos = isset($article['articleInfo']['array'])
			? array_map(function($info) {
				return [
					'id' => $info['infoId'],
					'type' => $info['infoType'],
					'name' => $info['infoTypeName'],
					'text' => preg_replace('#(https?://[^ ]+)#', '<a href="\\1" target="_blank">\\1</a>', $info['infoText']),
				];
			}, $article['articleInfo']['array'])
			: [];

		$oenumbers = isset($article['oenNumbers']['array'])
			? array_map(function($number) {
				return [
					'brand' => $number['brandName'],
					'number' => $number['oeNumber'],
				];
			}, $article['oenNumbers']['array'])
			: [];

		if ($this->attributesTable) {
            $attributeIds = array_unique(array_map(function($attribute) {
                return $attribute['id'];
            }, $attributes));

            if (!empty($attributeIds)) {
                $existantIds = array_map('intval', array_map(function($row) {
                    return $row['id'];
                }, $this->attributesTable
                    ->select(['id' => $attributeIds])
                    ->toArray()
                ));

                $insertPairs = [];

                foreach ($attributes as $attribute) {
                    if (!in_array($attribute['id'], $existantIds) && !isset($insertPairs[$attribute['id']])) {
                        $insertPairs[$attribute['id']] = $attribute['name'];
                    }
                }

                if (!empty($insertPairs)) {
                    foreach ($insertPairs as $id => $name) {
                        $this->attributesTable->insert([
                            'id' => $id,
                            'brand_no' => $article['directArticle']['brandNo'],
                            'article_no' => $article['directArticle']['articleNo'],
                            'name' => $name,
                        ]);
                    }
                }
            }
        }

		return [
			'id' => $article['directArticle']['articleId'],
			'genericArticleId' => $article['directArticle']['genericArticleId'],
			'key' => sprintf('%d|%s', $article['directArticle']['brandNo'], $article['directArticle']['articleNo']),
			'name' => $article['directArticle']['articleName'],
			'attributes' => $attributes,
			'infos' => $infos,
			'images' => $images,
			'brand' => [
				'id' => $article['directArticle']['brandNo'],
				'name' => $article['directArticle']['brandName'],
			],
			'oenumbers' => $oenumbers,
		];
	}

    /**
     * Vraća id artikla po ključu.
     *
     * @param $articleKey
     *
     * @return null
     *
     * @throws \Psr\Cache\InvalidArgumentException
     * @throws Exception
     */
	public function getArticleId($articleKey)
	{
		if (!$articleKey || preg_match('/^\d*\|$/', $articleKey)) {
			return null;
		}

        list($brandId, $articleNumber) = explode('|', $articleKey);

		if (preg_match('/^(39\d{2}|222)$/', $brandId)) {
		    return null;
        }

		$cacheKey = sprintf('articles.search.key.%s', md5($articleKey));
		$item = $this->cache->getItem($cacheKey);

		if (!$item->isHit()) {
			$articleKeyArray = explode('|', $articleKey);
			if (count($articleKeyArray) !== 2) return null;

			if (!$articleNumber) {
				var_dump($articleKey);
			}

			$articles = $this->request('getArticleDirectSearchAllNumbersWithState', [
				'brandId' => $brandId,
				'articleNumber' => $articleNumber,
				'numberType' => 10,
				'searchExact' => true,
			]);

			foreach ($articles as $article) {
				if ($article['articleNo'] == $articleNumber && $article['brandNo'] == $brandId) {
					$article = [
						'id' => $article['articleId'],
					];
					$item->set($article);
					$this->cache->save($item);
					break;
				}
			}
		}

		$article = $item->get();
		return $article ? $article['id'] : null;
	}

    /**
     * Vraća artikle po generic article id-evima.
     *
     * @param int[] $genericArticleIds
     *
     * @return array
     *
     * @throws \Psr\Cache\InvalidArgumentException
     * @throws Exception
     */
	public function getArticlesByGenericIds($genericArticleIds)
	{
		if (!($vehicle = $this->getSelectedVehicle())) {
			return [];
		}

		if (!is_array($genericArticleIds)) {
			$genericArticleIds = [$genericArticleIds];
		}

		sort($genericArticleIds);

		$cacheKey = sprintf('articles.vehicle.%d.generic.%s', $vehicle['id'], md5(implode(';', $genericArticleIds)));
		$item = $this->cache->getItem($cacheKey);

		if (!$item->isHit()) {
			$articleLinks = [];

			$chunks = array_chunk($genericArticleIds, 25);

			foreach ($chunks as $chunk) {
				$data = $this->request('getArticleIdsWithState', [
					'linkingTargetId' => $vehicle['id'],
					'genericArticleId' => ['array' => $chunk],
				]);

				$articleLinks = array_merge($articleLinks, array_map(function($article) {
					return [
						'articleId' => $article['articleId'],
						'articleLinkId' => $article['articleLinkId'],
					];
				}, $data));
			}

			$articles = $this->getArticlesByLinks($articleLinks);

			$item->set($articles);
			$this->cache->save($item);
		}

		return $item->get();
	}

    /**
     * Vrać artikle po assembly grupu.
     *
     * @param $assemblyGroupIds
     * @param int|null $linkingTargetId
     * @param array|null $brandNo
     * @param int $sort
     * @param string|null $linkingTargetType
     * @param string|null $articleCountry
     * @param string|null $lang
     * @param bool $returnArticleLinks
     *
     * @return array|mixed
     *
     * @throws \Psr\Cache\InvalidArgumentException
     * @throws Exception
     */
	public function getArticlesByAssemblyGroupIds(
	    $assemblyGroupIds,
        int $linkingTargetId = null,
        array $brandNo = null,
        int $sort = 1,
        string $linkingTargetType = null,
        string $articleCountry = null,
        string $lang = null,
        bool $returnArticleLinks = false
    ) {
		if ($linkingTargetId === null && !($vehicle = $this->getSelectedVehicle())) {
			return [];
		}

		if (null === $linkingTargetId) {
		    $linkingTargetId = $vehicle['id'];
        }

		if (!is_array($assemblyGroupIds)) {
			$assemblyGroupIds = [$assemblyGroupIds];
		}

		sort($assemblyGroupIds);

		$cacheKey = sprintf('articles.vehicle.%d.%s', $linkingTargetId, md5(implode(';', $assemblyGroupIds)));
		$item = $this->cache->getItem($cacheKey);

		if (false === $returnArticleLinks || !$item->isHit()) {
			$articleIds = [];
			$articleLinks = [];

            $params = [
                'linkingTargetId' => $linkingTargetId,
                'linkingTargetType' => $linkingTargetType,
                'articleCountry' => $articleCountry,
                'lang' => $lang,
                'sort' => $sort,
            ];

            if (is_array($brandNo)) {
                $params['brandNo'] = $brandNo;
            }

			foreach ($assemblyGroupIds as $assemblyGroupId) {
			    $params['assemblyGroupNodeId'] = intval($assemblyGroupId);

				$articleIdsForGroup = $this->request('getArticleIdsWithState', $params);
				$articleLinks = array_merge($articleLinks, $articleIdsForGroup);

				$articleIdsForGroup = array_map(function($article) {
					return $article['articleId'];
				}, $articleIdsForGroup);

				$articleIds = array_merge($articleIds, $articleIdsForGroup);
			}

			$articleIds = array_unique($articleIds);

            if ($returnArticleLinks) {
                return $articleLinks;
            }

			$articles = $this->getArticlesByIds($articleIds);

			$item->set($articles);
			$this->cache->save($item);
		}

		return $item->get();
	}

    /**
     * Vraća ključeve artikala po broju.
     *
     * @param string $articleNumber Broj artikla.
     * @param int|null $brandId
     * @param int|null $numberType
     * @param int $sortType
     * @param string|null $articleCountry
     * @param string|null $language
     * @param bool $returnArticles
     *
     * @return string[]
     *
     * @throws Exception
     */
    public function searchArticlesByNumber(
        string $articleNumber,
        int $brandId = null,
        int $numberType = null,
        int $sortType = 1,
        string $articleCountry = null,
        string $language = null,
        bool $returnArticles = false
    ): array {

	    $numberTypes = $numberType === null
            ? [0, 1, 3, 10]
            : [$numberType];

        $articles = [];

        $params = [
            'articleNumber' => $articleNumber,
            'articleCountry' => $articleCountry,
            'lang' => $language,
            'sortType' => $sortType,
            'searchExact' => true,
        ];

        if (null !== $brandId) {
            $params['brandId'] = $brandId;
        }

        foreach ($numberTypes as $number_type) {
            $params['numberType'] = $number_type;
            $articles = array_merge($articles, $this->request('getArticleDirectSearchAllNumbersWithState', $params));
        }

        if ($returnArticles) {
            return $articles;
        }

		return array_map(function($article) {
			return sprintf('%d|%s', $article['brandNo'], $article['articleNo']);
		}, $articles);
	}

	//</editor-fold>

    /**
     * @param int $linkingTargetId
     * @param int|null $parentNodeId
     * @param string|null $linkingTargetType
     * @param bool $childNodes
     * @param string|null $country
     * @param string|null $lang
     *
     * @return array|mixed
     * @throws Exception
     */
    public function getChildNodesAllLinkingTarget(
        int $linkingTargetId,
        int $parentNodeId = null,
        string $linkingTargetType = null,
        bool $childNodes = true,
        string $country = null,
        string $lang = null
    ) {
        return $this->request('getChildNodesAllLinkingTarget2', [
            'linkingTargetId' => $linkingTargetId,
            'parentNodeId' => $parentNodeId,
            'linkingTargetType' => $linkingTargetType,
            'childNodes' => $childNodes,
            'articleCountry' => $country,
            'lang' => $lang,
        ]);
    }

    /**
     * @return array|mixed
     * @throws Exception
     */
    public function getAmBrands()
    {
        return $this->request('getAmBrands');
    }

    /**
     * Vraća URL slike.
     *
     * @param $documentId
     * @param bool $thumbnail
     *
     * @return string
     */
	public function getImageUrl($documentId, $thumbnail = false)
	{
		return sprintf(
			'%s/pegasus-3-0/documents/%s/%s',
			preg_replace('#^[^/]+#', '', self::SERVICE_URI),
			$this->config['provider'],
			$documentId . ($thumbnail ? '/1' : '')
		);
	}

	/**
	 * Formatira godinu proizvodnje.
	 *
	 * @param string $year
	 * @param string $default
	 *
	 * @return string
	 */
    protected function formatConstructionYear($year, $default = '...')
    {
	    if (preg_match('/(\d{4})(\d{2})/', $year, $matches)) {
		    return sprintf('%s/%s', $matches[2], $matches[1]);
	    }

	    return $default;
    }
}
